        .386
        .model flat,stdcall

include win32.asi

        PUBLIC  qerr, ReadNumber, ReadAddress, Inputhandler
        PUBLIC  WadeSpace, GetInputLine, WadeSpaceOnly, InputBuffer
        PUBLIC  EraseLine, PutLine, EndOfLine

        extrn go : PROC, proceed : PROC, trace : PROC
        extrn PrintSpace : PROC, GetCurPos : PROC, SetCurPos : PROC
        extrn UnderlineCurs : PROC, BlockCurs : PROC
        extrn Dump : PROC, crlf : PROC, GetKey : PROC, WriteChar : PROC
        extrn WriteString : PROC, Unassemble : PROC
        extrn ModifyRegisters : PROC, Entry : PROC, Fill : PROC
        extrn Search : PROC, Move : proc, Compare : PROC, Math : PROC
        extrn ResetProcessor : PROC, Breaks : proc

IBSIZE = 80	;max size of iinput buffer for a line=width of screen

	.data

                          
InputBuffer db  IBSIZE+1 DUP (?)                  ;Allocate Input buffer
cursorbase dd   0
insert  db      0
prompt  db      13,10,"-> "

	.code
;
; Print an error if command wrong
;	Apparently, SI points to some location in the input buffer. We
; need to determine how many spaces to print before we found the actual
; error to point to from the line below. I guess if the error is in the
; first or second positions, we print the character right at the start of
; the line, else we space over to where the error is found.
;
qerr	PROC	
	call	crlf			;new line for this
        sub     esi,offset InputBuffer-2 ; Calculate error pos
        mov     ecx,esi                   ;
        jecxz    short qestart           ;
qelp:					
        call    PrintSpace              ; Space over to error pos
     	loop	qelp
qestart:
        mov     al,'^'                  ; Display error
        call    WriteChar
	stc				; Did an error
	ret	
qerr	ENDP	
;
; Read in a number.
;  In at least some cases, we have ALREADY called ReadReg to find a register,
;and failed so we are looking for a number anyway.
;	SI points to the input line from the user.
; OUTPUT: CY if we found 0 digits and did NOT find a register by name
;	EAX contains number found if NC
;
ReadNumber	PROC	
	push	ebx
        push    ecx
        push    edx
;        call    ReadReg                 ;find a non-number (mnemonic)
;        clc                             ;ignore which type or reg found
;        jz      gotnum                  ;if 0, we have value of this reg

;We reach here if we found no mnemonic in our lookup table

	sub	ebx,ebx			; Number = 0
	sub	cx,cx			; digits = 0
        mov     al,[esi]                 ;get next user character
	cmp	al,"'"			;if single or double quotes
	jz	getstring		;get a string input
	cmp	al,'"'
	jz	getstring
	cmp	al,'-'			;if a hyphen
        pushfd                           ;save result of compare
	jnz	rnlp			;not a hyphen, so SI is ok
        inc     esi                      ;else skip the hyphen
rnlp:
	lodsb				; Get char & convert to uppercase
	cmp	al,60h
	jc	notlower
	and	al,NOT 20h
notlower:
	sub	al,'0'    		; Convert to binary
	jc	short rn_done		; < '0' is an error
	cmp	al,10			; See if is a digit
	jc	short gotdigit		; Yes, got it
	sub	al,7			; Convert letters to binary
	cmp	al,16			; Make sure is < 'G'
	jnc	short rn_done		; Quit if not
	cmp	al,10			; MAke sure not < 'A'
	jc	short rn_done
gotdigit:
	shl	ebx,4			; It is a hex digit, add in
	or	bl,al			; OR in the digit
        inc     ecx                      ; Set flag to indicate we got digits
	jmp	rnlp			; go get the next digit

;We have snagged all entered digits here, and in the process SI got bumped
;past the non-digit by lodsb, so back up to look at the first non-digit
;character. the Zero flag contains the result of our hyphen compare, ZF if
;we found a hyphen, NZ if we didn't.
;  NOTE also that if we found a string, EBX has up to 4 ASCII codes in it,
;NOT a value. Don't offhand see any way the caller can know this, though.

rn_done:
        dec     esi                      ; Point at first non-digit
        popfd                            ;did we find a hyphen?
	jnz	rm_done2		; check for negation
	neg	ebx			;yes, negative, so negate value
rm_done2:
	mov	eax,ebx			;final result into EAX
	test	cl,-1			; See if got any CX=digit count
	jnz	gotnum			;if non-0, we got at least one
	stc				; No, error
gotnum:
        pop     edx
        pop     ecx
	pop	ebx
	ret				;return number/string in EAX

;We found a quote character at [SI], so we move past the quote to try to
;snag a quoted string. We keep sticking ASCII codes into EBX, shifting it
;left as we go, and losing all but the last 4 ASCII codes.

getstring:
        inc esi                          ;bump past quote character
getstringl:
	lodsb				;get next char
	cmp	al,13			;if CR, end of input
	jz	rm_done2		;so leave SI alone?
	cmp	al,"'"			;find end of quotes already?
	jz	rm_done2		;if so, check if null string
	cmp	al,'"'
	jz	rm_done2
	inc	cl    			;else we got a char
	shl	ebx,8			;stick in BL
	mov	bl,al
	jmp	getstringl		;and get the next one
ReadNumber	ENDP	
;
; Read an address, composed of a number and a possible selector
; INPUT: SI points at the input line just following the U (for unassemble).
; OUTPUT: EBX contains the address, which came either from a current
;	register bucket or from a string or number
;
ReadAddress	PROC	
	sub	edx,edx			;EDX used for segment if found
;        call    ReadReg                 ;chk for seg or non-seg register
					; mnemonic (like 'AX')

;ReadReg returns two flags: if NZ, we found nothing. if ZF, we found either
;a segment register or a non-segment register, so we examine the carry flag.
;CY means it was a non-segment register, and NC means we found a segment
;register.
;  If we found a register at all, the value returned is what was in the
;Regdump bucket for that register, and is returned in EAX unless it was not a
;32-bit register, in which case the low word is returned in AX and the high
;word of AX is zero. (undefined for segments)

;        jnc     gotseg                  ;if NC, we snagged a segment reg
;        mov     ebx,eax                 ;else, get value into ebx
;        jz      gotaddr                 ;ZF if non-segment reg found

;We fall through here if we failed to find a register. The syntax from the
;user permits either REG:number, REG:REG, number:REG, number:number, or just 
;number. At this point, no register was asked for by name, so we check for a 
;number.

readseg:
	call	ReadNumber              ; Read a number or string
	jc	short raerr		; Quit if error - no input
	mov	ebx,eax			;move number/string to EBX
	clc				; OK, exit
	ret
raerr:
	stc				; Error on number input
	ret
ReadAddress	ENDP	
;
; put a line to the screen
;
PutLine proc
        mov     eax,[cursorbase]
        call    SetCurPos
        mov     esi,offset InputBuffer
        mov     ecx,IBSIZE
pllp:
        lodsb
        push    eax
        call    WriteChar
        pop     eax
        
        or      al,al
        loopnz  pllp
        call    PrintSpace
        mov     eax,[cursorbase]
        mov     edx,edi
        sub     edx,offset InputBuffer
        add     eax,edx
        call    SetCurPos
        ret
PutLine endp
inquote PROC
        push    eax
        mov     bl,0
        mov     esi,offset InputBuffer
iql:
        lodsb
        or      al,al
        jz      iqx
        cmp     si,di
        ja      iqx
        cmp     al,bh
        jz      offquote
        cmp     al,'"'
        jz      onquote
        cmp     al,"'"
        jnz     iql
onquote:
        mov     bh,al
        inc     bl
        jmp     iql
offquote:
        sub     bx,bx
        jmp     iql

iqx:
        or      bl,bl
        pop     eax
        ret
inquote ENDP
EraseLine PROC
        pushfd
        mov     eax,[cursorbase]
        call    SetCurPos
        mov     esi,offset InputBuffer
elp:
        call    PrintSpace
        lodsb
        or      al,al
        jnz     elp
        mov     eax,[cursorbase]
        call    SetCurPos
        mov     edi,offset InputBuffer
        popfd    
        jnc      noclear
        mov     ecx,IBSIZE+1
        sub     al,al
        push    edi
        rep     stosb
        pop     edi
noclear:
        ret
EraseLine ENDP
EndOfLine PROC
        mov     edi,offset InputBuffer
        mov     esi,edi
        mov     ecx,-1
        mov     al,0
        repnz   scasb
        dec     edi
        neg     ecx
        dec     ecx
        push    ecx
        mov     eax,[cursorbase]
        add     eax,ecx
        dec     eax
        call    SetCurPos
        pop     ecx
        ret
EndOfLine ENDP
;
; Get an input line
; OUTPUT: SI points to the line of input.
;
GetInputLine	PROC	
        mov     [insert],1
        call    UnderlineCurs
        call    GetCurPos
        mov     [cursorbase],eax
        mov     byte ptr [InputBuffer],0
escape:
        stc
        call    EraseLine
moreinput:
	call	GetKey			;wait, return keystroke
        or      eax,eax
        jz      moreinput
        jns     asciichar
        cmp     ax,VK_LEFT
        jz      left
        cmp     ax,VK_RIGHT
        jz      right
        cmp     ax,VK_DELETE
        jz      delspc
        cmp     ax,VK_HOME
        jz      home
        cmp     ax,VK_END
        jz      doend
        cmp     ax,VK_INSERT
        jz      toginsert
        jmp     moreinput
asciichar :
        cmp     al,27                   ; esc - delete contents of line
        jz      escape
	cmp	al,9			; is tab?
	jz	dotab
	cmp	al,8			; Is delete or rubout?
        jz      bkspc             ; Yes - go do it
	cmp	al,7fh			;not on PC?
        jz      bkspc             ; yes - go do it
	cmp	al,13			; Is CR
        jz      endinput          ; Yes, return
        push    eax                      ;save keystroke
        call    inquote
        jnz     nolc                    ;if so, OK
	cmp	al,'A'			;see if below A
	jc	nolc
	cmp	al,'Z'			;or above Z
	ja	nolc
	or	al,20h			;if A-Z, force lowercase
nolc:
        test    [insert],1
        jz      overwrite
        push    edi
        mov     ebx,offset InputBuffer + IBSIZE - 1
        mov     esi,ebx
        sub     ebx,edi
        mov     ecx,ebx
        mov     edi,esi
        dec     esi
        std
        rep     movsb
        cld
        pop     edi
overwrite:
	stosb
        pop     eax
        call    PutLine
        cmp     edi,offset InputBuffer + IBSIZE
        jnz     moreinput
endinput:
        clc
        call    EraseLine
        call    PutLine
        call    EndOfLine
        mov     al,13
        stosb
	ret
toginsert:
        xor     [insert],1
        jnz     ulcurs
        call    BlockCurs
        jmp     moreinput
ulcurs:
        call    UnderlineCurs
        jmp     moreinput
home:
        mov     edi,offset InputBuffer
        mov     eax,[cursorbase]
        call    SetCurPos
        jmp     moreinput
doend:
        call    EndOfLine
        jmp     moreinput
bkspc:
        cmp     edi,offset InputBuffer
        jz      moreinput
        dec     edi
delspc:
        push    edi
        mov     ecx,offset InputBuffer + IBSIZE
        sub     ecx,edi
        mov     esi,edi
        inc     esi
        rep     movsb
        pop     edi
        call    PutLine
        jmp     moreinput
dotab:
        push    ecx
        mov     ecx,edi                   ; cx = current pos
        sub     ecx,esi                   ; Minus start pos
	add	cl,7			; now it equals number of
        and     ecx,7                    ; spaces per tab
        xor     ecx,7
	inc	cl			; spaces = 1 to 8
tablp:
	mov	al,' '
	stosb
        call    WriteChar
	loop	tablp
        pop     ecx
jmi:
	jmp	moreinput
left:
        cmp     edi,offset InputBuffer
        jz      jmi
        call    GetCurPos
        dec     ax
        call    SetCurPos
        dec     edi
        jmp     jmi
right:
        test    byte ptr [di],255
        jz      jmi
        inc     edi
        call    GetCurPos
        inc     ax
        call    SetCurPos
        jmp     jmi
GetInputLine	ENDP	
;
; Wade past spaces
; Whoa! By subtle implication, the carry accidentally gets set if the non-
; space of comma character was less than 13 in ASCII sequence.  This means
; if we hit a tab (8), the carry gets returned set.
;   Turns out this is NO ACCIDENT, and callers rely on this somehow.
;
WadeSpace       PROC    
	lodsb				; Get char
	cmp	al,' '			; if ' ' or ',' go again
        je      short WadeSpace         ;
	cmp	al,','			;
        je      short WadeSpace         ;
	cmp	al,9	 		;ignore tab also
        je      WadeSpace
        dec     esi                      ; Point at last space char
	cmp	al,13
	ret
WadeSpace       ENDP    
;
; Wade through spaces only
;
WadeSpaceOnly   PROC    
	lodsb			; Get a char
	cmp	al,' '		; Is space
        je      WadeSpaceOnly   ; Loop if so
	cmp	al,9		; ignore tab also
        je      WadeSpaceOnly
        dec     esi              ; Else point at char
	cmp	al,13
	ret
WadeSpaceOnly   ENDP 
;
; help routine
;
HelpText label byte
        db      13,10,"SIM-68K Version 1.00  Copyright (C) LADSoft (2002)",13,10,13,10
        db      "b [-]#,[addr]          breakpoints",13,10
        db      "c start, end, start2   compare memory",13,10
        db      "d [start [,end]]       dump memory",13,10
        db      "e [start] [value]      edit memory",13,10
        db      "f start,end, [fill]    fill memory",13,10
        db      "g [end]                go [until]",13,10
        db      "m start,end,start2     move memory",13,10
        db      "p                      proceed over call",13,10
        db      "q                      quit",13,10
        db      "r [name[ = value]]     display/modify reg",13,10
        db      "s start,end,value      search memory",13,10
        db      "t [count]              trace",13,10
        db      "u [start [,end]]       unassemble",13,10
        db      "x                      reset processor",13,10
        db      "?                      this text",13,10
        db      0
help PROC
        mov     edi,offset HelpText
        call    WriteString
        ret
help ENDP
;
; Main Input routine
;
Commands db     "bcdefghmpqrstux?"                     ; List of commands
comlen	= $ - Commands				; Length of list
cmdtab  label   dword
        dd      Breaks, Compare,Dump,Entry,Fill,go,Math
        dd      Move,proceed,quit,ModifyRegisters
        dd      Search,trace,Unassemble,ResetProcessor, help

Inputhandler    PROC
        mov     edi,offset prompt
        call    WriteString
	call	GetInputLine		; Get an input line
        call    WadeSpace               ; Wade through spaces       
        jz      Inputhandler            ; blank line, so try again
        inc     esi                      ; Point at first non-space char
        mov     edi,offset Commands      ; Get command list
        mov     ecx,comlen               ; Length of list
	repne	scasb			; search for command in list
	jnz	ierr			; Error if not in list
        mov     eax,comlen-1             ; Calculate position
        sub     eax,ecx                   ;
        call    [eax*4+cmdtab]
        jnc     Inputhandler            ; Get more input if no err
ierr:
	call	qerr			; Display error
        jmp     Inputhandler            ; Get more input
Inputhandler    ENDP
quit PROC
        pop     eax
        ret
quit ENDP
END